/********************************************
 * UTIA AV CR, v.v.i
 * Signal Processing Department
 * Pod Vodarenskou vezi 4
 * 182 08 Prague
 * Czech Republic
 * web: http://zs.utia.cas.cz
 * contact: sp@utia.cz
 *
 * Filename: clkwiz.c
 *
 * Date: 2022.03.07
 * Author: Lukas Kohout <kohoutl@utia.cas.cz>
 * Description:
 *    Implements functions for handling Video Timing Controller
 *
 * Changelog:
 *
 *
 *
 *
 ********************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
// #include <sleep.h>
// #include <xstatus.h>

#include "clkwiz.h"


// zynq 20 speed grade                     |   3     |   -2    | 1C/-1I/-1LI |  -1Q
// -------------------------------------------------------------------------------------
// MMCM_FVCOMIN Minimum MMCM VCO frequency | 600.00  | 600.00  | 600.00      | 600.00    MHz
// MMCM_FVCOMAX Maximum MMCM VCO frequency | 1600.00 | 1440.00 | 1200.00     | 1200.00   MHz


/*
 * table of predefined clkou0 parameters
 * calculator is located in internal/Vivado_2015_2/clkwiz_calc.zip
 *
 * clk_in = 50 MHz, VCO_RANGE = 600 - 1200 MHZ (ZYNQ speed grade 1C/-1I/-1LI/-1Q)
 */
mmcm_clkout0_t  clkout0_table_50_MHz[CLKWIZ_TAB_LEN] = {
//    clk_out   clk_div   clk_mul     clk_out_div
    { 148.50,   2,        37.125,     6.25    },
    { 128.44,   4,        51.375,     5.0     },
    { 135.00,   1,        23.625,     8.75    },
    { 119.23,   1,        23.25,      9.75    },
    { 108.00,   1,        20.25,      9.375   },
    {  74.25,   2,        37.125,     12.5    },
    {  65.00,   1,        22.75,      17.5    },
    {  40.00,   1,        23.5,       29.375  },
    {  25.19,   3,        58.0,       38.375  }
};


/*
 * clk_in = 100 MHz
 */
mmcm_clkout0_t  clkout0_table_100_MHz[CLKWIZ_TAB_LEN] = {
//    clk_out   clk_div   clk_mul     clk_out_div
    { 148.50,   4,        37.125,      6.250  },
    { 128.44,   5,        51.375,      8.000  },
    { 135.00,   2,        23.625,      8.750  },
    { 119.23,   2,        23.250,      9.750  },
    { 108.00,   2,        20.250,      9.375  },
    {  74.25,   4,        37.125,      12.50  },
    {  65.00,   1,        11.375,     17.500  },
    {  40.00,   1,        11.750,     29.375  },
    {  25.19,   5,        49.750,     39.500  }
};


/*
 * set particular register bits
 */
void clkwiz_set_reg32_bit(uintptr_t ctrl_base, unsigned int reg_base, unsigned int bit_mask)
{
    unsigned int reg = CLKWIZ_ReadReg32(ctrl_base, reg_base) | bit_mask;
    CLKWIZ_WriteReg32(ctrl_base, reg_base, reg);
}


/*
 * clear particular register bits
 */
void clkwiz_clear_reg32_bit(uintptr_t ctrl_base, unsigned int reg_base, unsigned int bit_mask)
{
    unsigned int reg = CLKWIZ_ReadReg32(ctrl_base, reg_base) & ~bit_mask;
    CLKWIZ_WriteReg32(ctrl_base, reg_base, reg);
}


/*
 * get clock wizard status register
 */
inline unsigned int clkwiz_get_status(uintptr_t ctrl_base)
{
    return CLKWIZ_ReadReg32(ctrl_base, CLKWIZ_STATUS_OFFSET);
}


/*
 * get clock wizard lock flag
 */
inline unsigned int clkwiz_is_locked(uintptr_t ctrl_base)
{
    unsigned int status = CLKWIZ_ReadReg32(ctrl_base, CLKWIZ_STATUS_OFFSET);
    return (status & CLKWIZ_LOCKED) ? 1 : 0;
}


/*
 * reset clock wizard
 */
void inline clkwiz_reset (uintptr_t ctrl_base)
{
    clkwiz_set_reg32_bit(ctrl_base, CLKWIZ_SW_RESET_OFFSET, 0x0000000A);
}


/*
 * set VCO frequency
 */
void clkwiz_set_vco(clkwiz_t *clkwiz)
{
    unsigned int  reg = 0;
    float         clk_mul_int;
    float         clk_mul_frac = modff(clkwiz->clkout0.clk_mul, &clk_mul_int);

    reg =  (unsigned int)clkwiz->clkout0.clk_div;       // Bit[7:0]    = DIVCLK_DIVIDE
    reg |= (unsigned int)clk_mul_int << 8;              // Bit[15:8]   = CLKFBOUT_MULT_INT
    reg |= (unsigned int)(clk_mul_frac * 1000.0) << 16; // Bit[25:16]  = CLKFBOUT_MULT_FRAC
    reg |= 0x04000000;                                  // Bit[26]     = CLKFBOUT_FRAC_EN

//     printf("VCO_REG 0x%08X\n", reg);

    CLKWIZ_WriteReg32(clkwiz->base_addr, CLKWIZ_VCO_OFFSET, reg);
}


/*
 * configure clkout0 output to generate required frequency
 */
void clkwiz_set_clkout0(clkwiz_t *clkwiz)
{
    unsigned int  reg = 0;
    float         clk_out_div_int;
    float         clk_out_div_frac = modff(clkwiz->clkout0.clk_out_div, &clk_out_div_int);

    reg =  (unsigned int)clk_out_div_int;                   // Bit[7:0]    = CLKOUT0_DIVIDE
    reg |= (unsigned int)(clk_out_div_frac * 1000.0) << 8;  // Bit[17:8]  = CLKOUT0_FRAC
    reg |= 0x00040000;                                      // Bit[18]     = CLKFBOUT_FRAC_EN

//     printf("CLKOUT0_REG 0x%08X\n", reg);

    CLKWIZ_WriteReg32(clkwiz->base_addr, CLKWIZ_CLKOUT0_DIVIDE_OFFSET, reg);
}


/*
 * set clock wizard clkout0 parameters from the table
 */
void clkwiz_clkout0_lookup(mmcm_clkout0_t *clkout0, unsigned char sel, float clk_in)
{
    mmcm_clkout0_t *clkout0_table;

    assert( (clk_in == 50.0f) || (clk_in == 100.0f) );

    if ( clk_in == 50.0f ) {
        clkout0_table = clkout0_table_50_MHz;
    } else if ( clk_in == 100.0f ) {
        clkout0_table = clkout0_table_100_MHz;
    } else {
        clkout0_table = clkout0_table_50_MHz;
    }

    clkout0->clk_out      = clkout0_table[sel].clk_out;
    clkout0->clk_div      = clkout0_table[sel].clk_div;
    clkout0->clk_mul      = clkout0_table[sel].clk_mul;
    clkout0->clk_out_div  = clkout0_table[sel].clk_out_div;
}


/*
 * initialize clock wizard to generate selected frequency on clkout0
 */
int clkwiz_set_clock(clkwiz_t *clkwiz, unsigned char clk_sel)
{
    int status;
    int timeout;

    clkwiz_clkout0_lookup(&(clkwiz->clkout0), clk_sel, clkwiz->clk_in);

    clkwiz_set_vco(clkwiz);
    clkwiz_set_clkout0(clkwiz);
    clkwiz_update(clkwiz->base_addr);

    // wait on lock - stable output clock
    timeout = 100;
    status = 0;
    while (  !status && timeout-- ) {
        status = clkwiz_is_locked(clkwiz->base_addr);
        if (!status) usleep(10);
    }
    if (!status) return -1;

    return 0;
}
